#!/bin/bash

# Some formatting
RED="\033[0;31m"
NC="\033[0m" # No Color

RUNDIR=$(pwd)
PARAM_FILE=${RUNDIR}/param.cfg

# Source parameter with config paths
if [[ ! -f ${PARAM_FILE} ]]
then
   echo -e "No param file "${PARAM_FILE} "found\n${RED}Abort ...${NC}"
   exit 1
else 
  echo "Sourcing param files"
  source ${PARAM_FILE}

  ADDITIONAL_FILES_DIR=${RUNDIR}/Utils

  # Check if given path exists
  if [[ ! -d ${DWARF_HOME} ]]
  then
    echo -e "The specified folder does not exist: "${DWARF_HOME}"\n${RED}Abort${NC}"
    exit 1 
  else
    DWARF_SRC=${DWARF_HOME}/src
    DWARF_NAMELIST=${DWARF_HOME}/escape-acraneb2-input-output/nam_radia_dwarf.nam
  fi
fi

# Folder which contains files for RPE compiling
DWARF_RPE=${RUNDIR//Local}RPE_IMPLEMENTATION
DWARF_RPE_HOME=${DWARF_RPE}/${DWARF_HOME##*/}
DWARF_RPE_SRC=${DWARF_RPE_HOME}/${DWARF_SRC##*/}
DWARF_RPE_NAMELIST=${DWARF_RPE_HOME}/${DWARF_NAMELIST##*${DWARF_HOME}/}

#download last AutoRPE version
#git clone http://earth.bsc.es/gitlab/otinto/AutoRPE.git --branch dwarf-lonlev-0.91_0

AUTORP_HOME=${RUNDIR}/AutoRPE

if [[ ! -d ${AUTORP_HOME} ]]
then
  echo -e "AutoRPE folder not found in : "${AUTORP_HOME}"\n${RED}Abort${NC}"
  exit 1
else 
  # Paths hardcoded in Implement_sources
  AUTO_RP_IN=${AUTORP_HOME}/ModelSources/Input/
  AUTO_RP_OUT=${AUTORP_HOME}/ModelSources/Output/

  # Paths harcoded in NamelistPrecisionsGenerator script
  OLD_NAMELIST=${AUTORP_HOME}/ParameterNamelists/Original/
  NEW_NAMELIST=${AUTORP_HOME}/ParameterNamelists/Modified/
fi
# Remove eventual leftovers of old runs
rm -rf   ${DWARF_RPE} ${AUTO_RP_IN}/* ${AUTO_RP_OUT}/*

mkdir ${DWARF_RPE}


# ---- Start running the emulator ---- 

# Copy .f90 files
cp ${DWARF_SRC}/*.f90 ${AUTO_RP_IN}

# Precompile and copy F90 files
for file in ${DWARF_SRC}/*.F90
do
 dest_file=${file//${DWARF_SRC}}
 cpp -P -fopenmp  ${file} > ${AUTO_RP_IN}/${dest_file//F90}f90
done

mapfile -t check_reals < <(grep -rin -w '^\s*real\s*' ${AUTO_RP_IN} | grep -v '::' | grep -v -i 'function')

if [[ ${#check_reals[@]} ]]
then
  echo -e "${RED}Found some REALS declared without '::' ${NC}in files:"
  for files  in "${check_reals[@]}"
  do
    echo ${files##*/}
  done
  echo -e "For the moment we will try to manually fix it \nIf something goes wrong change the declaration properly ..." 
  for files  in "${check_reals[@]}"
  do
    file=`echo $check_reals | awk -F':' '{print $1}'`
    line=`echo $check_reals | awk -F':' '{print $2}'`
    var=`echo $files | awk -F':' '{print $3}'`
    new_var=${var/real/real ::}
    new_var=${var/REAL/REAL ::}
    sed -i "${line}s/^\s*${var}/${new_var}/" ${file} || exit 1
    echo -e "${RED}Succeed ${NC}in changing declaration of var "${var} "to "${new_var}
  done
fi

# Remove legacy files
rm ${AUTO_RP_IN}/*_legacy*

# Ignore opmpi files and timings, which contains no calculation at all
echo -e  \
 'dmi_omp.F90' \
 '\nini_modules.f90' \
 '\ninput_acraneb2.f90' \
 '\noutput_basic.f90' \
 '\ndmi_omp.f90' \
 '\ndmi_timer.f90' > ${AUTORP_HOME}/AdditionalFiles/list_of_files_to_keep_unmodified.txt


# Check python is loaded
python3 -V  >/dev/null 2>&1  ||  { echo -e "${RED}load python3 and try again ...${NC}"; exit 1 ; }

cd ${AUTORP_HOME}

# Run the implementation
python3 ImplementRPEtoSource.py 2> implementation.log

# Add call to read variables precisions to mail file 
sed -i "s/ call domp_init(nthreads)/ call domp_init(nthreads)\n CALL read_variable_precisions/g" ${AUTO_RP_OUT}/main.f90 

#Copy processed files
echo "Copying files to "${DWARF_RPE}" folder ..."
cp -r ${DWARF_HOME} ${DWARF_RPE} || exit 1
cp ${AUTO_RP_OUT}/* ${DWARF_RPE_SRC} || exit 1

# Add the new dependecies for compiling
cp ${ADDITIONAL_FILES_DIR}/dependencies ${DWARF_RPE_SRC}/.dependencies.in || exit 1
cp ${ADDITIONAL_FILES_DIR}/objs_spec ${DWARF_RPE_SRC}/.objs.spec.in || exit 1

# Copy source file of rp_emulator
cp  ${ADDITIONAL_FILES_DIR}/rp_emulator.F90 ${DWARF_RPE_SRC} || exit 1

# Copy dependecies of rp_emulator.F90
cp -r ${ADDITIONAL_FILES_DIR}/include  ${DWARF_RPE} || exit 1

# Add options for compiling and configuring 

# Add the free length line to the compiling option
FCFLAGS='-ffree-line-length-none -fopenmp -g -O2 -cpp'
sed -i  "/configure / s/$/ FCFLAGS=\'${FCFLAGS}\'/" ${DWARF_RPE_HOME}/confmake.sh

# Add include folder for dependencies of rp_emulator
INC_FOLDER=${DWARF_RPE}/include/
INC_FOLDER='MKLINCL:=-I'${INC_FOLDER}''
sed -i 's@^MKLINCL.*@'${INC_FOLDER}'@' ${DWARF_RPE_HOME}/Makefile.include.in

# Add make clean to confmake
sed -i  "s/^make /make clean\nmake/" ${DWARF_RPE_HOME}/confmake.sh

# Work with Namelist 

# Copy namelist original namelist
cp ${DWARF_NAMELIST} ${OLD_NAMELIST} || exit 1

# Generate namelist_precisions
python3 NamelistPrecisionsGenerator.py

# Modify original namelist
python3 AdaptNamelists.py

# Copy the namelists in the RPE folder
cp ${NEW_NAMELIST}/* ${DWARF_RPE_NAMELIST} || exit 1
cp namelist_precisions  ${DWARF_RPE_NAMELIST} || exit 1

echo "All files has been copied"

# ---- Start compiling section
cd ${DWARF_RPE_HOME}

module load gcc/7.2.0

# Compile
./confmake.sh

echo -e "Executable produced\n"${DWARF_RPE_HOME//src}"/dwarf"
